'''
Author: Qirong Li
Date: February 2019
Project: Row reduction game
'''

import pygame
import constants as cs
import math, random

class Game:
    def __init__(self, board_size=4):

        if board_size > 14 or board_size < 2:
            print('board size must be between 2 and 14')
            return

        self.board_size = board_size
        self.block_size = (cs.grid_right - cs.grid_left)//board_size - 2
        self.radius = int(self.block_size/3)
        

    def init(self):

        self.grid = self.init_grid()
        self.buttons = self.init_buttons()
        self.texts = self.init_texts()
        self.step = 0
        self.chosen = []
        self.start_time = pygame.time.get_ticks()

    def start(self):

        pygame.init()
        pygame.display.set_caption("game")
        self.screen = pygame.display.set_mode([cs.window_size, cs.window_size])

        self.init()
        clock = pygame.time.Clock()
        pygame.mixer.music.load('rv_cj.mp3')
        pygame.mixer.music.play(-1)

        done = False
        solved = False
        
        while not done:

            self.screen.fill(cs.grey)

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    done = True
                    break
                if event.type == pygame.MOUSEBUTTONDOWN:
                    pos = pygame.mouse.get_pos()
                    self.click_button(pos)

                    if self.is_goal():
                        self.texts[0][3] = 'Congratulations! You solved the puzzle!'
                        solved = True

                if not solved:
                    self.texts[1][3] = 'Time: ' + str(int((pygame.time.get_ticks() - self.start_time)/1000)) + ' s, ' + \
                        'Steps: ' + str(self.step)
                self.draw(self.screen)
                
                # pygame.display.update()
                pygame.display.flip()
                clock.tick(60)

        pygame.quit()

    def is_goal(self):

        count = 0

        for row in range(self.board_size):
            for col in range(self.board_size):

                color = self.grid[row][col][2]
                # the original goal
                if color == 0:
                    return False
                elif color == 1:
                    count += 1

                # if color != -1:
                #     count += 1

        return (count <= self.board_size)

    def click_button(self, pos):

        px, py = pos
        clicked = -1
        for i in range(len(self.buttons)):
            x, y, color, text, width, height = self.buttons[i]
            if x <= px <= x + width and y <= py <= y + height:
                clicked = i
                # print('clicked: ', clicked)
                break

        # select
        if 0 <= clicked < self.board_size:

            if not self.chosen:
                self.chosen.append(clicked)
                self.buttons[clicked] = [x, y, cs.green, '1', width, height]

            elif len(self.chosen) == 1:
                if clicked == self.chosen[0]:
                    self.buttons[clicked][3] = '1 & 2'
                else:
                    self.buttons[clicked] = [x, y, cs.green, '2', width, height]
                self.chosen.append(clicked)

        # cancel
        elif self.board_size <= clicked < 2*self.board_size:

            clicked = clicked - self.board_size
            x, y, color, text, width, height = self.buttons[clicked]

            if len(self.chosen) == 1 and clicked == self.chosen[0]:
                self.buttons[clicked] = [x, y, cs.milk_blue, 'select', width, height]
                self.chosen = []

            elif len(self.chosen) == 2:
                if clicked == self.chosen[0]:
                    if self.chosen[0] == self.chosen[1]:
                        self.buttons[clicked] = [x, y, cs.milk_blue, 'select', width, height]
                        self.chosen = []
                    else:
                        self.buttons[clicked] = [x, y, cs.milk_blue, 'select', width, height]
                        prev2 = self.chosen[1]
                        self.buttons[prev2][3] = '1'
                        self.chosen = [prev2]
                elif clicked == self.chosen[1]:
                    self.buttons[clicked] = [x, y, cs.milk_blue, 'select', width, height]
                    self.chosen = [self.chosen[0]]

        # swap
        elif clicked == 2*self.board_size:

            if len(self.chosen) != 2:

                self.texts[0][3] = 'please select exactly two rows'

            else:
                self.swap_rows(self.chosen[0], self.chosen[1])
                self.texts[0][3] = ''
                self.step += 1

        # add
        elif clicked == 2*self.board_size + 1:

            if len(self.chosen) != 2:
                if not self.texts:
                    self.texts.append([])
                self.texts[0][3] = 'please select exactly two rows'

            else:
                self.add_rows(self.chosen[0], self.chosen[1])
                self.texts[0][3] = ''
                self.step += 1

        # new game
        elif clicked == 2*self.board_size + 2:

            self.init()

    def add_rows(self, row1, row2):
        # add rows
        # -1: empty, 0: white, 1: black 
        for i in range(self.board_size):
            c1, c2 = self.grid[row1][i][2], self.grid[row2][i][2]
            if c1 == -1:
                continue
            elif c1 == c2:
                self.grid[row2][i][2] = 1 - c2
            elif c2 == -1:
                self.grid[row2][i][2] = c1
            else: # 1 black 1 white
                self.grid[row2][i][2] = -1
        
        # change buttons
        self.buttons[self.chosen[0]][2] = cs.milk_blue
        self.buttons[self.chosen[0]][3] = 'select'
        self.buttons[self.chosen[1]][2] = cs.milk_blue
        self.buttons[self.chosen[1]][3] = 'select'
        self.chosen = []


    def swap_rows(self, row1, row2):
        # swap rows
        for i in range(self.board_size):
            self.grid[row1][i][2], self.grid[row2][i][2] = self.grid[row2][i][2], self.grid[row1][i][2]
        
        # change buttons
        self.buttons[self.chosen[0]][2] = cs.milk_blue
        self.buttons[self.chosen[0]][3] = 'select'
        self.buttons[self.chosen[1]][2] = cs.milk_blue
        self.buttons[self.chosen[1]][3] = 'select'
        self.chosen = []

    # [x, y, color]
    # -1: empty, 0: white, 1: black 
    def init_grid(self):

        grid = []

        for y in range(self.board_size):
            grid.append([])
            for x in range(self.board_size):
                if x == y: # diagonal: black
                    color = 1
                else:
                    rand = random.uniform(0, 1)

                    if rand <= 1/3:
                        color = 1
                    else:
                        color = 0

                grid[y].append([cs.grid_left + x* (self.block_size + 2), \
                    cs.grid_top + y* (self.block_size + 2), color])
        # print(grid)
        return grid

    # [x, y, color, string, width, height]
    def init_buttons(self):

        buttons = []
        diff = cs.window_size-cs.grid_right
        for y in range(self.board_size):
            buttons.append([cs.grid_right+int((1/6)*diff), \
                    cs.grid_top + y*(self.block_size+2)+12, cs.milk_blue, 'select', int(0.28*diff), 25])
        
        for y in range(self.board_size):
            buttons.append([cs.grid_right+int((1/2)*diff), \
                    cs.grid_top + y*(self.block_size+2)+12, cs.milk_blue, 'cancel', int(0.28*diff), 25])

        buttons.append([int(0.25 * cs.window_size), int(cs.grid_bottom + 0.3 * (cs.window_size - cs.grid_bottom)), \
            cs.milk_blue, 'swap', 60, 30])

        buttons.append([int(0.5 * cs.window_size), int(cs.grid_bottom + 0.3 * (cs.window_size - cs.grid_bottom)), \
            cs.milk_blue, 'add', 60, 30])

        buttons.append([int(0.75 * cs.window_size), int(cs.grid_bottom + 0.3 * (cs.window_size - cs.grid_bottom)), \
            cs.milk_blue, 'new game', 80, 30])

        return buttons

    def init_texts(self):

        texts = [[int(0.5 * cs.window_size), int(0.3 * cs.grid_top), cs.pink, ''], \
        [int(0.5 * cs.window_size), int(0.6 * cs.grid_top), cs.pink, 'Time: ' + str(0) + ' s']]
        return texts

    def draw(self, screen):

        for row in self.grid:
            for block in row:
                x, y, color = block

                rect = pygame.Rect(x, y, self.block_size, self.block_size)
                pygame.draw.rect(screen, cs.wood, rect)

                if color == 0:
                    pygame.draw.circle(screen, cs.white, (x + int(0.5*self.block_size), \
                        y + int(0.5*self.block_size)), self.radius, 0)
                elif color == 1:
                    pygame.draw.circle(screen, cs.black, (x + int(0.5*self.block_size), \
                        y + int(0.5*self.block_size)), self.radius, 0)

        for button in self.buttons:

            x, y, color, text, width, height = button
            # rect
            rect = pygame.Rect(x, y, width, height)
            pygame.draw.rect(screen, color, rect)
            # text
            textSurf, textRect = text_object(text, cs.black)
            textRect.center = (x + int(width/2), y + int(height/2))
            screen.blit(textSurf, textRect)

        for t in self.texts:
            
            x, y, color, text = t
            textSurf, textRect = text_object(text, color)
            textRect.center = (x, y)

            screen.blit(textSurf, textRect)

def text_object(text, color):
    # font = pygame.font.Font("freesansbold.ttf",15)
    font = pygame.font.SysFont("comicsansms", 20)
    textSurface = font.render(text, True, color)
    return (textSurface, textSurface.get_rect())

